/*
 * main.cpp
 *
 *  Created on: 2021年12月20日
 *      Author: ubuntu
 */

#include <iostream>
#include <nvdstracker.h>
#include <fstream>
#include <zconf.h>

#include <opencv2/videoio.hpp>
#include <opencv2/imgproc.hpp>
#include <driver_types.h>
#include <cuda_runtime_api.h>
#include<opencv2/highgui.hpp>

#ifdef DYNAMIC_LOAD
#include <dlfcn.h>

typedef NvMOTStatus (*NV_Query)(uint16_t customConfigFilePathSize, char* pCustomConfigFilePath, NvMOTQuery *pQuery);
typedef NvMOTStatus (*NV_Init)(NvMOTConfig *pConfigIn,
                               NvMOTContextHandle *pContextHandle,
                               NvMOTConfigResponse *pConfigResponse);
typedef NvMOTStatus (*NV_Process)(NvMOTContextHandle contextHandle,
                                  NvMOTProcessParams *pParams,
                                  NvMOTTrackedObjBatch *pTrackedObjectsBatch);
#endif

NvMOTFrame                    _nvFrame;
NvMOTProcessParams            _nvProcessParams;
NvBufSurface                * _nvBufSurface;
NvMOTTrackedObjList         * _nvTrackedObjList;
NvMOTTrackedObjBatch          _nvTrackedObjBatch;
NvMOTObjToTrack             * _nvObjsToTrack;
NvMOTObjToTrackList         * _nvObjsToTrackList;
NvMOTContextHandle             contextHandle;

int trackerMaxCount = 30;
auto width = 640;
auto height = 512;

int main()
{
    auto deepstreamPath = "/opt/nvidia/deepstream/deepstream-"+std::string("5.1");
    auto trackerCustomConfigFile = std::string(deepstreamPath + "/sources/apps/sample_apps/deepstream-test2/tracker_config.yml");
    auto trackerCustomConfigFileRecasted = const_cast<char *>(trackerCustomConfigFile.c_str());
    auto trackerLibLocation = std::string(deepstreamPath + "/lib/" + "libnvds_mot_klt"+ ".so");

    std::string trackerConfigFile = "nvidia_tracker_config.txt";

    {
        std::ofstream outfile;
        outfile.open(trackerConfigFile, std::ofstream::trunc);
        outfile << "[tracker]\r\n";
        outfile << "tracker-width=" + std::to_string(width) + "\r\n";
        outfile << "tracker-height=" + std::to_string(height) + "\r\n";
        outfile << "gpu-id=0\r\n";
        outfile << "ll-lib-file=" + trackerLibLocation + "\r\n";
        outfile << "ll-config-file=" + trackerCustomConfigFile + "\r\n";
        outfile << "enable-batch-process=0\r\n";
        outfile.flush();
        outfile.close();
        usleep(1000000 / 2);
    }

#ifdef DYNAMIC_LOAD
    auto _libHandle = dlopen(trackerLibLocation.c_str(),
                                    RTLD_NOW | RTLD_GLOBAL);

    auto _queryFunction = (NV_Query) dlsym(_libHandle, "NvMOT_Query");
    auto _initFunction = (NV_Init) dlsym(_libHandle, "NvMOT_Init");
    auto _processFunction = (NV_Process)dlsym(_libHandle, "NvMOT_Process");
#endif

    NvMOTQuery query;
#ifdef DYNAMIC_LOAD
    _queryFunction(trackerConfigFile.size(),
                   const_cast<char *>(trackerConfigFile.c_str()),
                   &query);
#else
    NvMOT_Query(trackerConfigFile.size(),
                const_cast<char *>(trackerConfigFile.c_str()),
                &query);
#endif

    NvBufSurfaceCreateParams surfaceCreateParams;
    surfaceCreateParams.layout = NVBUF_LAYOUT_PITCH;
    surfaceCreateParams.colorFormat = NVBUF_COLOR_FORMAT_RGBA;
    surfaceCreateParams.width = width;
    surfaceCreateParams.height = height;
    surfaceCreateParams.isContiguous = false;
    surfaceCreateParams.size = 0;
    surfaceCreateParams.gpuId = 0;
    surfaceCreateParams.memType = query.memType;

    _nvBufSurface = new NvBufSurface();
    auto surfaceCreateResult = NvBufSurfaceCreate(&_nvBufSurface, 1, &surfaceCreateParams);
    if (surfaceCreateResult == -1)
    {
        std::cout << "NvBufSurfaceCreate failed with value " << surfaceCreateResult << std::endl;
        return 0;
    }
    NvBufSurfaceMemSet(_nvBufSurface, -1, -1, 0);
    auto nvSurfaceMapResult = NvBufSurfaceMap(_nvBufSurface, 0, -1,
                                              NvBufSurfaceMemMapFlags::NVBUF_MAP_READ_WRITE);


    NvMOTPerTransformBatchConfig transformBatchConfig;
    transformBatchConfig.bufferType = query.memType;
    transformBatchConfig.colorFormat = NVBUF_COLOR_FORMAT_RGBA;
    transformBatchConfig.maxWidth = _nvBufSurface->surfaceList->width;
    transformBatchConfig.maxHeight = _nvBufSurface->surfaceList->height;
    transformBatchConfig.maxPitch = _nvBufSurface->surfaceList->pitch;
    transformBatchConfig.maxSize = _nvBufSurface->surfaceList->dataSize;

    NvMOTMiscConfig miscConfig;
    miscConfig.gpuId = 0;
    miscConfig.maxObjPerBatch = 0;
    miscConfig.maxObjPerStream = 0;

    NvMOTConfig config;
    config.computeConfig = query.computeConfig;
    config.maxStreams = 1;
    config.customConfigFilePath = trackerCustomConfigFileRecasted;
    config.customConfigFilePathSize = trackerCustomConfigFile.size();
    config.perTransformBatchConfig = &transformBatchConfig;
    config.numTransforms = query.numTransforms;
    config.miscConfig = miscConfig;

    NvMOTConfigResponse configResponse;

#ifdef DYNAMIC_LOAD
    auto status_init = _initFunction(&config, &contextHandle, &configResponse);
#else
    auto status_init = NvMOT_Init(&config, &contextHandle, &configResponse);
#endif

    if(status_init == NvMOTStatus_OK) {
        std::cout << "STATUS INIT OK" << std::endl;
    } else {
        std::cout << "STATUS INIT FAILED" << std::endl;
        return 0;
    }

    {
        _nvFrame.streamID = 0;
        _nvFrame.doTracking = true;
        _nvFrame.reset = true;
        _nvFrame.timeStamp = 0; // Set on incoming frames
        _nvFrame.timeStampValid = true;
        _nvFrame.numBuffers = 1;
        _nvFrame.bufferList = &_nvBufSurface->surfaceList;
        _nvFrame.objectsIn.numFilled = 1;
        _nvFrame.objectsIn.detectionDone = false;
        _nvFrame.objectsIn.numAllocated = 1;
        _nvFrame.objectsIn.list = nullptr;
    }

    {
        _nvProcessParams.numFrames = 1;
        _nvProcessParams.frameList = &_nvFrame;
    }

    {
        _nvTrackedObjList = new NvMOTTrackedObjList[1];
        _nvTrackedObjList->list = new NvMOTTrackedObj[trackerMaxCount];
        _nvTrackedObjList->numAllocated = trackerMaxCount;
        _nvTrackedObjList->numFilled = 0;
        _nvTrackedObjList->streamID = 0;
        _nvTrackedObjList->frameNum = 0;
        _nvTrackedObjList->valid = true;

        _nvTrackedObjBatch.list = _nvTrackedObjList;
        _nvTrackedObjBatch.numFilled = 1;
        _nvTrackedObjBatch.numAllocated = 1;
        _nvObjsToTrack = new NvMOTObjToTrack[trackerMaxCount];

        _nvObjsToTrackList = new NvMOTObjToTrackList();
        _nvObjsToTrackList->list = _nvObjsToTrack;
        _nvObjsToTrackList->numAllocated = trackerMaxCount;
        _nvObjsToTrackList->detectionDone = false;
    }

    cv::VideoCapture _videoCapture_appearance("/opt/nvidia/deepstream/deepstream-5.1/samples/streams/sample_720p.h264");

    while(true)
    {
        cv::Mat image;
        auto image_read = _videoCapture_appearance.read(image);
        if (image_read)
        {
#ifdef DCF_TRACKER
            cv::cvtColor(image, image, cv::COLOR_BGR2RGBA);
#else
            cv::cvtColor(image, image, cv::COLOR_BGR2RGB);
#endif

            cv::resize(image, image, cv::Size(width, height));

            {
                if(_nvFrame.objectsIn.list == nullptr)
                {
                    _nvFrame.objectsIn.list = new NvMOTObjToTrack[trackerMaxCount];

					_nvFrame.objectsIn.list[0].bbox = NvMOTRect { 185, 220, 30,30 };
					_nvFrame.objectsIn.list[0].doTracking = true;
					_nvFrame.objectsIn.list[0].classId = 0;
					_nvFrame.objectsIn.list[0].confidence = -0.1f;
					_nvFrame.objectsIn.numFilled = 1;
					_nvFrame.objectsIn.numAllocated = trackerMaxCount;
					_nvFrame.objectsIn.detectionDone = false;
                }
            }

            // Populate Frame
            {
                cudaMemcpy(_nvBufSurface->surfaceList->mappedAddr.addr[0], image.data,
                           _nvBufSurface->surfaceList->dataSize, cudaMemcpyHostToDevice);

            }

#ifdef DYNAMIC_LOAD
            auto status_process = _processFunction(contextHandle, &_nvProcessParams, &_nvTrackedObjBatch);
#else
            auto status_process = NvMOT_Process(contextHandle, &_nvProcessParams, &_nvTrackedObjBatch);
#endif

            if (status_process == NvMOTStatus_Error)
            {
                std::cout << "Could not process frame!" << std::endl;
            }
            else
            {
                for (int i = 0; i < _nvTrackedObjBatch.numFilled; i++)
                {
                    for (int j = 0; j < _nvTrackedObjBatch.list[i].numFilled; j++)
                    {
                        auto currentTrackedObj = _nvTrackedObjBatch.list[i].list[j];
                        std::cout << " i: "<< i<< " j: "<< j<< " Confidence: "<< currentTrackedObj.confidence << " Found Tracker: " << currentTrackedObj.trackingId
                        		<<" Ref [x: "<< currentTrackedObj.bbox.x<<" y: "<< currentTrackedObj.bbox.y
                        		<<" w: "<< currentTrackedObj.bbox.width<<" h: "<< currentTrackedObj.bbox.height<<"]"<<std::endl;
                        cv::Rect rect = cv::Rect(currentTrackedObj.bbox.x,currentTrackedObj.bbox.y,currentTrackedObj.bbox.width,currentTrackedObj.bbox.height);
                        cv::rectangle(image, rect, cv::Scalar(0, 0, 255.0));
                    }
                }
            }
            _nvFrame.frameNum += 1;
            _nvFrame.doTracking = true;
            _nvFrame.reset = false;
            if(_nvFrame.frameNum == 100){
            	_nvFrame.reset = true;
            }

        }
        cv::imshow("image",image);
        cv::waitKey(30);
//        usleep(1000000 / 30);
    }

    return 0;
}


